import base64
from typing import Optional
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import sentinel

from fred_api.registry.keyset.keyset_history_types_pb2 import (
    KeysetHistoryRecord,
    KeysetHistoryReply,
    KeysetHistoryRequest,
)
from fred_api.registry.keyset.keyset_info_types_pb2 import (
    DnsKey as DnsKeyMessage,
    KeysetIdReply,
    KeysetIdRequest,
    KeysetInfoReply,
    KeysetInfoRequest,
)
from fred_api.registry.keyset.keyset_state_flags_types_pb2 import KeysetStateFlagsReply
from fred_api.registry.keyset.keyset_state_history_types_pb2 import (
    KeysetStateHistory,
    KeysetStateHistoryReply,
    KeysetStateHistoryRequest,
)
from fred_api.registry.keyset.keyset_state_types_pb2 import KeysetStateReply, KeysetStateRequest
from fred_types import KeysetHistoryId, KeysetId
from frgal.utils import AsyncTestClientMixin, make_awaitable

from regal.keyset import DnsKey, DSRecord, Keyset, KeysetClient
from regal.tests.common import ObjectClientTestMixin


class TestKeysetClient(AsyncTestClientMixin, KeysetClient):
    """Test KeysetClient."""


class KeysetClientTest(ObjectClientTestMixin[TestKeysetClient], IsolatedAsyncioTestCase):
    """Test KeysetClient's methods."""

    grpc_namespace = "/Fred.Registry.Api.Keyset.Keyset"
    get_state_request = KeysetStateRequest
    get_state_reply = KeysetStateReply
    get_state_method = "get_keyset_state"
    get_history_request = KeysetHistoryRequest
    get_history_reply = KeysetHistoryReply
    get_history_method = "get_keyset_history"
    get_history_record = KeysetHistoryRecord
    get_state_history_request = KeysetStateHistoryRequest
    get_state_history_reply = KeysetStateHistoryReply
    get_state_history_method = "get_keyset_state_history"
    get_state_history_record = KeysetStateHistory.Record
    get_state_flags_method = "get_keyset_state_flags"
    get_state_flags_reply_cls = KeysetStateFlagsReply

    def setUp(self):
        self.client = TestKeysetClient(netloc=sentinel.netloc)

    async def test_get_keyset_id(self):
        keyset_id = KeysetId("kryten")
        reply = KeysetIdReply()
        reply.data.keyset_id.uuid.value = str(keyset_id)
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_keyset_id("HANDLE")
        self.assertEqual(response, keyset_id)
        self.assertIsInstance(response, KeysetId)

        request = KeysetIdRequest()
        request.keyset_handle.value = "HANDLE"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Keyset.Keyset/get_keyset_id",
            timeout=None,
        )

    async def _test_get_keyset_info(self, keyset_history_id: Optional[KeysetHistoryId] = None) -> None:
        dns_key = DnsKeyMessage()
        dns_key.flags = 256
        dns_key.protocol = 3
        dns_key.alg = 15
        dns_key.key = "some key"
        reply = KeysetInfoReply()
        reply.data.keyset_id.uuid.value = "kryten"
        if keyset_history_id:
            reply.data.keyset_history_id.uuid.value = keyset_history_id
        reply.data.keyset_handle.value = "HANDLE"
        reply.data.dns_keys.extend([dns_key])
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_keyset_info(KeysetId("kryten"), history_id=keyset_history_id)

        self.assertEqual(
            response,
            Keyset(
                keyset_id=KeysetId("kryten"),
                keyset_history_id=keyset_history_id,
                keyset_handle="HANDLE",
                dns_keys=[DnsKey(flags=256, protocol=3, alg=15, key="some key")],
            ),
        )

        request = KeysetInfoRequest()
        request.keyset_id.uuid.value = "kryten"
        if keyset_history_id:
            request.keyset_history_id.uuid.value = keyset_history_id
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Keyset.Keyset/get_keyset_info",
            timeout=None,
        )

    async def test_get_keyset_info(self):
        await self._test_get_keyset_info()

    async def test_get_keyset_info_with_history_id(self):
        await self._test_get_keyset_info(keyset_history_id=KeysetHistoryId("nova-5"))


class DnsKeyTest(TestCase):
    def test_ds_record(self):
        data = (
            (
                "example.com",
                DnsKey(
                    flags=257,
                    alg=13,
                    key="kXKkvWU3vGYfTJGl3qBd4qhiWp5aRs7YtkCJxD2d+t7KXqwahww5IgJtxJT2yFItlggazyfXqJEVOmMJ3qT0tQ==",
                ),
                DSRecord(
                    keytag=370,
                    alg=13,
                    digest_type=2,
                    digest="BE74359954660069D5C63D200C39F5603827D7DD02B56F120EE9F3A86764247C",
                ),
            ),
            # Example from RFC 4509, section 2.3
            # See https://datatracker.ietf.org/doc/html/rfc4509#section-2.3
            (
                "dskey.example.com",
                DnsKey(
                    flags=256,
                    alg=5,
                    key="AQOeiiR0GOMYkDshWoSKz9XzfwJr1AYtsmx3TGkJaNXVbfi/2pHm822aJ5iI9BMzNXxeYCmZDRD99WYwYqUSdjMmmAphXdvxegXd/M5+X7OrzKBaMbCVdFLUUh6DhweJBjEVv5f2wwjM9XzcnOf+EPbtG9DMBmADjFDc2w/rljwvFw==",
                ),
                DSRecord(
                    keytag=60485,
                    alg=5,
                    digest_type=2,
                    digest="D4B7D520E7BB5F0F67674A0CCEB1E3E0614B93C4F9E99B8383F6A1E4469DA50A",
                ),
            ),
            # Example of high algorithm number
            (
                "example.com",
                DnsKey(
                    flags=257,
                    alg=254,
                    key="Z2F6cGFjaG8h",
                ),
                DSRecord(
                    keytag=53667,
                    alg=254,
                    digest_type=2,
                    digest="65D86A34355F375FE8A4EA41F4619CCB52499EDA07CEBCF6D66DE4916101279C",
                ),
            ),
        )

        for fqdn, dnskey, dsrecord in data:
            with self.subTest(fqdn=fqdn, dnskey=dnskey, dsrecord=dsrecord):
                self.assertEqual(dnskey.ds_record(fqdn), dsrecord)

    def test_keytag_md5(self):
        data = (
            (DnsKey(flags=256, alg=1, key=base64.b64encode(b"abcdefgh").decode()), 25958),
            (DnsKey(flags=256, alg=1, key=base64.b64encode(b"abc").decode()), 0),
        )

        for dnskey, keytag in data:
            with self.subTest(dnskey=dnskey, keytag=keytag):
                self.assertEqual(dnskey._ds_record_keytag(), keytag)
